#!/usr/bin/python
# Mlt2edl is an altered version of Kdenparse.
# Alternations by Tin2tin, 2014
# Kdenparse is
# Copyright 2011 Will Riley
# Distributed under the terms of the GNU General Public License v3

"""
 About kdenparse.py:
 
 KdenParse attempts to extract and present useful information 
 from a .kdenlive XML project file to a Premiere friendly EDL file.


 About mltparse.py:

 mltparse attempts to extract and present useful information
 from a .mlt (Shotcut) to a Blender friendly EDL file.

 Needs the Timecode lib: https://pypi.python.org/pypi/timecode

 Important:
 Source clips must not contain spaces and can only be max 8 chr long not counting the file extension.
 Blender uses the reel tag to identify clips (not the from clip: line.)

 Done:
 - ms to fr conversion.
 - added Timecode lib
 - added black clips and disabled it again - using space instead.
 - added 0 to hours string
 - added clip name (max 8 chr - no space= in reel name position.
 - transitions (dissolve)
 - save edl to file.

 Unfinished:
 - Fade In/out/Wipe - very complicated to extract on the MLT side.
 - MLT hide/mute - but Blender EDL only works with "a" - audio only or "v" - video only tracks or else crash.
 - Blender: Video with audio "va"/"b" makes it crash. Blender black clips missing audio?
 - Blender: why are clips from the same track sometimes added to other tracks?

 Blender import problems:
 - VA/va/B/b does not add audio track. Requested a fix in the Blender EDL import script: http://lists.blender.org/pipermail/bf-vfx/2014-November/000447.html
 - AA/V is commonly used by ex. Vegas and Lightworks, but is not supported by the Blender import script.
 - Missing format options in Blender: AA/V for Sony Vegas/Lightworks - AAV Avid? AA stereo. REM used as alternate for * or #
 - Not searching for reels in the same folder as the edl file is located.
 - Not using 'From Clip Name:' instead of reel info.

 Usage: "mltparse.py --edl "my.mlt" will save "my.edl" in the same folder as the original file.

"""
import os, sys, argparse
from timecode import Timecode

version_ = "%(prog)s 0.1.0"

argparser = argparse.ArgumentParser(prog="mlt2edl.py",
    add_help=True,
    description="Parses .mlt project files for various metadata and time line informations.")
argparser.add_argument('-V', '--version', 
    action='version', 
    version=version_)
argparser.add_argument('--edl', action='store_true', default=False,
    dest='create_edl',
    help='Generate EDL output.')
argparser.add_argument('--deref-proxy', action='store_true', default=False,
    dest='deref_proxy',
    help='Dereference proxy clip (show original filenames)')
argparser.add_argument('--frames', action='store_true', default=False,
    dest='show_frames',
    help='Show frames instead of TC when using --edl.')
argparser.add_argument('--links', action='store_true', default=False,
    dest='show_links',
    help='Show source id to filename association.')
argparser.add_argument('--profile', action='store_true', default=False,
    dest='get_profile',
    help='Generate project profile metadata.')
argparser.add_argument('--producers', action='store_true', default=False,
    dest='get_producers',
    help='Show MLT producers (media file) metadata.')
argparser.add_argument('--tractors', action='store_true', default=False,
    dest='get_tractors',
    help='Show MLT tractors (media file) metadata.')
argparser.add_argument('--kproducers', action='store_true', default=False,
    dest='get_kproducers',
    help='Show Kdenlive producers (media file) metadata.')
argparser.add_argument('projectFile')

args = argparser.parse_args()

if not os.path.isfile(args.projectFile):
    print "Not a file we can work with..."
    sys.exit(1)

try: 
    args.projectFile.rindex(".mlt",-9)
except ValueError:
    print "Invalid filename. Exiting."
    sys.exit(1)


from xml.dom import minidom
from decimal import Decimal,getcontext,ROUND_DOWN
from math import modf, floor

class KdenParse:
        
    def __init__(self, kdenliveFile):
        self.xmldoc = minidom.parse(kdenliveFile)

    def getProjectProfile(self):
        profileDict = {}
        profile = self.xmldoc.getElementsByTagName("profile")
        keyList = profile.item(0).attributes.keys()
        for a in keyList:
            profileDict[a] = profile.item(0).attributes[a].value
        return profileDict


    def msfr(self, fr, timestring):
        if len(timestring) == 12:
            frs = float(timestring[9:12])
            frs=float(frs/1000)*float(fr)
            frs = "%02d" % (round(frs))
            timestring=timestring[0:8]+":"+frs
        return timestring

    def getXml(self, url):
        try:
            print url
            f = urllib2.urlopen(url)
        except:
            f = url
        print f
        doc = xml.dom.minidom.parse(f)
        node = self.documentElement
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            print 'Element name: %s' % node.nodeName
            for (name, value) in node.attributes.items():
                #print '    Attr -- Name: %s  Value: %s' % (name, value)
                if name == 'reminder':
                    self.rem_value = value

    def getTracks(self):
        tracks = []
        t = self.xmldoc.getElementsByTagName("tracks")
        for track in t:
            tracks.append(track.attributes["producer"].value) 
        return tuple(tracks)
    
    def getPlaylists(self):
        playlistList = []
        playlists = self.xmldoc.getElementsByTagName("playlist")
        projectMeta = self.getProjectProfile()
        framerate = float(str(float(projectMeta["frame_rate_num"]) / float(projectMeta["frame_rate_den"]))[0:5])
        for p in playlists:#[::-1]:
            eventList = []
            plDict = {}
            plDict["pid"] = p.attributes["id"].value
            plDict["format"] = "V"
            formatevent = p.getElementsByTagName("property")
            for fe in formatevent:
                if fe.attributes["name"].value=="shotcut:audio":
                    plDict["format"] = "A"#fe.attributes["name"].value
                elif fe.attributes["name"].value=="shotcut:video":
                    plDict["format"] = "AA/V"#change to AA/V for Sony Vegas/Lightworks - AAV Avid?

            events = p.getElementsByTagName("*")
            for event in events:
                if event.hasAttribute("length"):
                    evDict = {}
                    evDict["producer"] = "black"
                    evDict["inTime"] = "00:00:00:00"
                    evDict["outTime"] = str(Timecode(framerate,self.msfr(framerate,event.attributes["length"].value))- Timecode(framerate,"00:00:00:00"))#MLTblacks are 1 frame longer than "out"
                    evDict["Trans"] = "C"
                    eventList.append(evDict)
                if event.hasAttribute("producer") and event.attributes["producer"].value[0:7]=="tractor":
                    tractors = self.xmldoc.getElementsByTagName("tractor")
                    for t in tractors:
                        if str(t.attributes["id"].value)==str(event.attributes["producer"].value):
                            tDict = {}
                            tes = t.getElementsByTagName("track")
                            cnt=0
                            for te in tes:
                                if cnt == 0:
                                    dDict = {}
                                    dDict["producer"] = te.attributes["producer"].value.replace(" ","_")
                                    dDict["inTime"] = te.attributes["in"].value
                                    dDict["outTime"] = te.attributes["in"].value
                                    dDict["Trans"] = "C"
                                    eventList.append(dDict)

                                if cnt == 1:
                                    tDict = {}
                                    tDict["producer"] = te.attributes["producer"].value.replace(" ","_")
                                    tDict["inTime"] = te.attributes["in"].value
                                    tDict["outTime"] = te.attributes["out"].value
                                    tDict["Trans"] = "D "+str(Timecode(framerate,self.msfr(framerate,(t.attributes["out"].value))).frames)
                                    eventList.append(tDict)
                                cnt=cnt+1

                if event.hasAttribute("producer") and event.attributes["producer"].value[0:7]!="tractor" and event.attributes["producer"].value!="black":
                    evDict = {}
                    evDict["producer"] = event.attributes["producer"].value.replace(" ","_")
                    evDict["inTime"] = event.attributes["in"].value
                    evDict["outTime"] = event.attributes["out"].value
                    evDict["Trans"] = "C"
                    eventList.append(evDict)
            plDict["events"] = eventList
            playlistList.append(plDict)
        return tuple(playlistList)
    
    def getKProducers(self):
        kProducerList = []
        profile = self.xmldoc.getElementsByTagName("kdenlive_producer")
        for i in profile:
            kpDict = {}
            keyList = i.attributes.keys()
            for a in keyList:
                kpDict[i.attributes[a].name] = i.attributes[a].value
                print kpDict[i.attributes[a].name]
            kProducerList.append(kpDict)
        return tuple(kProducerList)
    
    def getProducers(self):
        producerList = []
        producers = self.xmldoc.getElementsByTagName("producer")
        for p in producers:
            pDict = {}
            pDict["pid"] = p.attributes["id"].value
            pDict["inTime"] = p.attributes["in"].value
            pDict["outTime"] = p.attributes["out"].value
            properties = p.getElementsByTagName("property")
            for props in properties:
                try:
                    pDict[props.attributes["name"].value.replace(".","_")] = props.firstChild.data
                except AttributeError:
                    pass
                # FIX FOR ATTIBUTE ERROR
                # https://kldp.org/node/151698
                #pDict[props.attributes["name"].value.replace(".","_")] = props.firstChild.data
            producerList.append(pDict)
        return tuple(producerList)

    def getTractors(self):
        tractorList = []
        tractors = self.xmldoc.getElementsByTagName("tractor")
        for t in tractors:
            tDict = {}
            tDict["pid"] = t.attributes["id"].value
            tDict["inTime"] = t.attributes["in"].value
            tDict["outTime"] = t.attributes["out"].value
            properties = t.getElementsByTagName("property")
            for props in properties:
                tDict[props.attributes["name"].value.replace(".","_")] = props.firstChild.data
            tractorList.append(tDict)
        return tuple(tractorList)

    def linkReferences(self):
        sourceLinks = {}
        for i in self.getProducers():
            srcPid = i["pid"]
            sourceLinks[srcPid] = i["resource"]
        return sourceLinks
    
    def derefProxy(self):
        proxyLinks = {}
        for i in self.getKProducers():
            try:
                if i["proxy"]:
                    _proxy = i["proxy"]
                    proxyLinks[_proxy] = i["resource"]
            except KeyError:
                return False
        return proxyLinks

    def createEdl(self):
        projectMeta = self.getProjectProfile()
        framerate = float(str(float(projectMeta["frame_rate_num"]) / float(projectMeta["frame_rate_den"]))[0:5])
        sourceLinks = self.linkReferences()
        EDLfile=""
        for playlist in self.getPlaylists():
            if playlist["pid"]!="main bin" and playlist["pid"]!="background":
                EdlEventCnt = 1
                progIn = Timecode(framerate,'00:00:00:00') # showtime tally
                progOut = Timecode(framerate,'00:00:00:00')
                progIn.frame_number == 0
                progOut.frame_number == 0
                srcChannel = "C" # default channel/track assignment
                srcType=playlist["format"]+"     "[0:4-len(playlist["format"])]#Exact 5 chr wide
                EDLfile+= "\n === " + playlist["pid"] + " === \n\n"
                for event in playlist["events"]:
                    prod = event["producer"]
                    #prodChunks = prod.split("_")
                    #srcType = prodChunks[-1].capitalize()[:1]
                    # if it's an audio event, extract channel info from producer id
                    #if srcType == "A":
                    #   srcChannel = prodChunks[1]
                    srcIn = self.msfr(framerate, event["inTime"].replace('.', ':')) # source clip IN time
                    srcIn = Timecode(framerate,srcIn)
                    #srcInfr= srcIn.frames
                    srcOut = self.msfr(framerate, event["outTime"].replace('.', ':')) # source clip OUT time
                    srcOut = Timecode(framerate,srcOut)
                    #srcOutfr= srcOut.frames
                    #if EdlEventCnt == 1: #why? Blender needs out==in
                    #    progOut = progOut - Timecode(framerate,"00:00:00:00")
    #                srcDur = srcOutfr - srcInfr
                    #progOut = progOut + (srcOut - srcIn) # increment program tally
                    progOut = progOut + (srcOut - srcIn)#-Timecode(framerate,"00:00:00:00")
                    sourcePath = sourceLinks[prod]
                    sourceFile = sourcePath.split("/")[-1]
                    proxyList = self.derefProxy()
                    if proxyList:
                        try:
                            sourceRef = "(" + proxyList[sourcePath].split("/")[-1] + ") " + sourceFile
                        except KeyError:
                            sourceRef = sourceFile
                    else:
                        sourceRef = sourceFile

                    if sourceRef == "black":
                        #newsourceRef="BL"#remove insert of black clips
                        progIn = (progIn + (srcOut - srcIn))
                    else:
                        EDLfile+= "* FROM CLIP NAME: " + sourceRef+"\n"
                        fnl=len(sourceRef)-4
                        newsourceRef=sourceRef[0:fnl][fnl-8:fnl] #add 8 chr of filename w. no extention. NB! no space in filename.
                        newsourceRef=newsourceRef+"         "[0:8-len(newsourceRef)]#9 chr wide
                        if event["Trans"][0:1]=="D":
                            EdlEventCnt=EdlEventCnt-1

                        EEC = "%03d" % (int(EdlEventCnt))
                        EDLfile+= str(EEC) + "  " + newsourceRef + " "
                        #print str(EdlEventCnt) + "  " + prod + "  ", #alt. use producername.
                        EDLfile+= srcType + " " + event["Trans"] + "         "[0:10-len(event["Trans"])]

                        if args.show_frames:
                            EDLfile+=  str(srcIn.frames) + " " + str(srcOut.frames) + ""
                            EDLfile+=  str(progIn.frames) + " " + str(progOut.frames)
                        else:
                            EDLfile+=  str(srcIn) + " " + str(srcOut) + " "
                            EDLfile+=  str(progIn) + " " + str(progOut)+"\n"
                            #print self.framesToDF(srcIn) + " " + self.framesToDF(srcOut) + "", #timecode lib replaces framesToDF
                            #print self.framesToDF(progIn) + " " + self.framesToDF(progOut) #timecode lib replaces framesToDF

                        #if EdlEventCnt == 1:
                        #    progIn = progIn + Timecode(framerate,"00:00:00:00")
                        #progIn = progIn + Timecode(framerate,"00:00:00:00")
                        progIn = (progIn + (srcOut - srcIn))
                        EdlEventCnt = EdlEventCnt+1
        print EDLfile
        with open(args.projectFile[0:len(args.projectFile)-4]+".edl", 'w') as f:
            f.write(EDLfile)
            f.close
            print "Saved: "+args.projectFile[0:len(args.projectFile)-4]+".edl"

    def framesToABS(self, frameAddress):#obsolete - using Timecode instead
        getcontext().prec = 10
        getcontext().rounding = ROUND_DOWN
        projectMeta = self.getProjectProfile()
        frameRate = Decimal(projectMeta["frame_rate_num"]) / Decimal(projectMeta["frame_rate_den"])
        frameDuration = 1 / frameRate
        #print "fps = " + str(frameRate)
        #print "1 fr = " + str(frameDuration)  + " secs"
        absDuration = Decimal(frameAddress) * Decimal(frameDuration) # frameAddress length in seconds
        #print "TC = " + str(absDuration)
        f, w = modf(absDuration) # split float at decimal (fraction, whole)
        #print "Split: " + "%f + %f" % (w, f)
        frameRemainder = Decimal(str(f)) / Decimal(frameDuration)
        m, s = divmod(w, 60)
        h, m = divmod(m, 60)
        tc = "%02d:%02d:%02d:%02d" % (h, m, s, frameRemainder)
        return tc
        
    def framesToDF(self, framenumber):#obsolete - using Timecode instead.
        """
            This method adapted from C++ code called "timecode" by Jason Wood.
            begin: Wed Dec 17 2003
            copyright: (C) 2003 by Jason Wood
            email: jasonwood@blueyonder.co.uk 
            Framerate should be 29.97, 59.94, or 23.976, otherwise the calculations will be off.
        """

        projectMeta = self.getProjectProfile()
        framerate = float(projectMeta["frame_rate_num"]) / float(projectMeta["frame_rate_den"])
        
        # Number of frames to drop on the minute marks is the nearest integer to 6% of the framerate
        dropFrames = round(framerate * 0.066666) 
        # Number of frames in an hour
        framesPerHour = round(framerate * 60 * 60) 
        # Number of frames in a day - timecode rolls over after 24 hours
        framesPerDay = framesPerHour * 24 
        # Number of frames per ten minutes
        framesPer10Minutes = round(framerate * 60 * 10) 
        # Number of frames per minute is the round of the framerate * 60 minus the number of dropped frames
        framesPerMinute = (round(framerate) * 60) - dropFrames 
        
        if (framenumber < 0): # For negative time, add 24 hours.
            framenumber = framesPerDay + framenumber

        # If framenumber is greater than 24 hrs, next operation will rollover clock
        # % is the modulus operator, which returns a remainder. a % b = the remainder of a/b

        framenumber = framenumber % framesPerDay 
        d = floor(framenumber / framesPer10Minutes)
        m = framenumber % framesPer10Minutes

        if (m > 1):
            framenumber=framenumber + (dropFrames * 9 * d) + dropFrames * floor((m-dropFrames) / framesPerMinute)
        else:
            framenumber = framenumber + dropFrames * 9 * d;

        frRound = round(framerate);
        frames = framenumber % frRound;
        seconds = floor(framenumber / frRound) % 60;
        minutes = floor(floor(framenumber / frRound) / 60) % 60;
        hours = floor(floor(floor(framenumber / frRound) / 60) / 60);    

        tc = "%02d:%02d:%02d:%02d" % (hours, minutes, seconds, frames)#added two zeros in hours
        return tc

    def framesToNDF(self):
        pass

kp = KdenParse(args.projectFile)

if args.create_edl:
    kp.createEdl()

if args.get_profile:
    for i in kp.getProjectProfile().keys():
        print i + ": " + kp.getProjectProfile()[i]    

if args.get_tractors:
    for i in kp.getTractors():
        print "\n=================\n"
        for kv in i:
            print kv + ": " + i[kv]

if args.get_producers:
    for i in kp.getProducers():
        print "\n=================\n"
        for kv in i:
            print kv + ": " + i[kv]

if args.get_kproducers:
    for i in kp.getKProducers():
        print "\n=================\n"
        for kv in i:
            print kv + ": " + i[kv]
        
if args.deref_proxy:
    for i in kp.derefProxy():
        print i + ": " + kp.derefProxy()[i]
        
if args.show_links:
    for i in kp.linkReferences():
        print i + ": " + kp.linkReferences()[i]

